! ********************************************** 
!   09/01/2008 
! ********************************************** 
! 
! Computes the roots of a given polynomial P 
! applying the BEGG fast QR algorithm to the 
! associated companion matrix A. 
! 
!
! Here the update step for V and F is cheaper 
! than in previous versions because it is only 
! partially performed. Also,  
! all arrays of rank two or three 
! have been replaced by arrays of rank one or
! "simple" variables.
!
! In this version the Lapack routine ZGEEV is applied
! to a larger matrix (of size LAPACKSIZE).
!
! **************************************************************
!
! Input variables:
!
! N            degree of P 
!
! POLY         complex array of length N, containing the nonleading 
!              coefficients of P (ordered with decreasing degree)
!
! LAPACKSIZE   size of residual matrix whose eigenvalues are
!              computed using Lapack (please choose LAPACKSIZE>3) 
!
! Output variables:
!
! MYEIG        complex array of length N, containing the roots
!              of P.
!
! ***************************************************************


subroutine begg_eig_lapack(N,POLY,LAPACKSIZE,MYEIG)

   
  implicit none 

  integer, intent(in) :: n,LAPACKSIZE
  complex(kind(1d0)), intent(in) :: poly(n) 
  complex(kind(1d0)), intent(out) :: myeig(n) 
  complex(kind(1d0)), allocatable :: work(:),auxAA(:,:),vl(:,:),vr(:,:),eig(:)
  complex(kind(1d0)) :: Q11(n-1),Q12(n-1),Q21(n-1),Q22(n-1),p(n),q(n),sigma(n-1)
  complex(kind(1d0)) :: b11(n-1),b12(n-1),b21(n-1),b22(n-1) 
  complex(kind(1d0)) :: V11(n-1),V12(n-1),V21(n-1),V22(n-1) 
  complex(kind(1d0)) :: Vtau11(n-1),Vtau12(n-1),Vtau21(n-1),Vtau22(n-1),Vhat33(n-1) 
  complex(kind(1d0)) :: Vhat11(n-1),Vhat12(n-1),Vhat13(n-1),Vhat21(n-1),Vhat22(n-1),Vhat23(n-1),Vhat31(n-1),Vhat32(n-1) 
  complex(kind(1d0)) :: F11(n-1),F12(n-1),F13(n-1),F21(n-1),F22(n-1),F23(n-1),F31(n-1),F32(n-1),F33(n-1) 
  complex(kind(1d0)) :: newV11(n-1),newV12(n-1),newV21(n-1),newV22(n-1),newF33(n-1) 
  complex(kind(1d0)) :: newF11(n-1),newF12(n-1),newF13(n-1),newF21(n-1),newF22(n-1),newF23(n-1),newF31(n-1),newF32(n-1) 
  complex(kind(1d0)) :: Ftau11(n-2),Ftau12(n-2),Ftau13(n-2),Ftau21(n-2),Ftau22(n-2),Ftau23(n-2),Ftau31(n-2),Ftau32(n-2) 
  complex(kind(1d0)) :: Fhat11(n-2),Fhat12(n-2),Fhat13(n-2),Fhat14(n-2),Fhat21(n-2),Fhat22(n-2),Fhat23(n-2),Fhat24(n-2) 
  complex(kind(1d0)) :: Fhat31(n-2),Fhat32(n-2),Fhat33(n-2),Fhat34(n-2),Fhat41(n-2),Fhat42(n-2),Fhat43(n-2),Fhat44(n-2) 
  complex(kind(1d0)) :: g1(n),g2(n),h1(n),h2(n),gamma1(n-1),gamma2(n-1),Ftau33(n-2) 
  complex(kind(1d0)) :: newp(n),newq(n),ptau(n),qtau(n),sigmatau(n),gtau1(n),gtau2(n) 
  complex(kind(1d0)) :: diagonals(n),subdiag(n-1),beta(n) 
   
  complex(kind(1d0)) :: Giv11,Giv12,Giv21,Giv22,x(2),y(2),alpha,auxA12,auxA13,auxA21,auxA22,auxA23,auxA31,auxA32,auxA33 
  complex(kind(1d0)) :: bulge11,bulge12,bulge21,bulge22,z(2) 
  complex(kind(1d0)) :: auxB11,auxB12,auxB21,auxB22,XX12,XX13,XX14,WW11,WW12,WW21,WW22,WW23,WW31,WW32,WW33 
  complex(kind(1d0)) :: ZZ11,ZZ12,ZZ21,ZZ22,auxM11,auxM12,auxM13,auxM21,auxM22,auxM23,auxC21,auxC22,auxC23 
  complex(kind(1d0)) :: MM11,MM12,MM13,MM22,MM23,auxMM11,auxMM12,auxMM13,auxMM21,auxMM22,auxMM23 
  complex(kind(1d0)) :: XXX11,XXX12,XXX21,XXX22,XXX31,XXX32,YYY12,YYY22,YYY31,YYY32 
  complex(kind(1d0)) :: auxBB21,auxBB22,auxBB31,auxBB32,DD11,DD12,DD13,DD14,DD21,DD22,DD23,DD24,DD31,DD32,DD33,DD34,DD42,DD43,DD44 
  real(kind(1d0)), allocatable :: rwork(:)
  real(kind(1d0)) :: w1,w2,normA,threshold,our_time,maxdiffeR,maxdiffeI

  integer :: iterations(n) 
  integer :: clock_start,clock_end,clock_rate 
  integer :: i,j,k,kp1,kp2,m,nm1,nm2,defl_size,random_flag,repeat_flag,lwork,info 

CALL SYSTEM_CLOCK(COUNT_RATE=clock_rate) 
  CALL SYSTEM_CLOCK(COUNT=clock_start) 


  !************* SETUP ************************* 
   
   
  normA=sqrt(sum(poly*conjg(poly))+n-1) 
  !threshold=1.d0*epsilon(normA) 
  threshold=normA*epsilon(1.d0) 
  !write(*,*) "Norm(A) = ",normA 
  !write(*,*) "Threshold = ",threshold 
  defl_size=n-1; 
  do i=1,n 
     iterations(i)=0 
  enddo 
   
  random_flag=0  ! flag for random shift 

  !lapacksize=5  ! usual value is 4
   
  nm1=n-1 
  nm2=n-2 
   
  !************************************************* 
  ! Initialization of F, V, p, q. 
  !************************************************* 
   
  ! Store permuted identity matrices in F: 
  do i=1,(n-3) 
     F11(i)=0.d0 
     F13(i)=0.d0 
     F22(i)=0.d0 
     F23(i)=0.d0 
     F31(i)=0.d0 
     F32(i)=0.d0 
     F21(i)=(1.d0,0.d0) 
     F12(i)=(1.d0,0.d0) 
     F33(i)=(1.d0,0.d0) 
  enddo 
  F11(nm2)=0.d0 
  F12(nm2)=0.d0 
  F22(nm2)=0.d0 
  F23(nm2)=0.d0 
  F31(nm2)=0.d0 
  F33(nm2)=0.d0 
  F13(nm2)=(1.d0,0.d0) 
  F21(nm2)=(1.d0,0.d0) 
  F32(nm2)=(1.d0,0.d0) 
   
  F12(nm1)=0.d0 
  F13(nm1)=0.d0 
  F21(nm1)=0.d0 
  F23(nm1)=0.d0 
  F31(nm1)=0.d0 
  F32(nm1)=0.d0 
  F11(nm1)=(1.d0,0.d0) 
  F22(nm1)=(1.d0,0.d0) 
  F33(nm1)=(1.d0,0.d0) 
   
  ! Store identity matrices in V: 
  do i=1,(nm1) 
     V12(i)=0.d0 
     V11(i)=(1.d0,0.d0) 
     V22(i)=(1.d0,0.d0) 
     V21(i)=0.d0 
  enddo 
   
  q=poly; 
  q(n)=q(n)+(1.d0,0.d0) 
  p=0.d0 
  p(1)=(1.d0,0.d0) 
   
   
  !******* BEGIN APPROX CYCLE ************************ 
   
  approx:do while (defl_size>(lapacksize-2)) 
     ! Initialize Q with identity matrices. 
      
     do i=1,(nm1) 
        Q12(i)=0.d0 
        Q11(i)=(1.d0,0.d0) 
        Q22(i)=(1.d0,0.d0) 
        Q21(i)=0.d0 
     enddo 
      
     ! Use the previously computed V and F to update the 
     ! quasiseparable representation (step 1 in Thm. 3.1). 
      
     h1(1:nm2)=F11(1:nm2) 
     h2(1:nm2)=F21(1:nm2)  
      
     b11(1:nm2)=F12(1:nm2) 
     b12(1:nm2)=F13(1:nm2) 
     b21(1:nm2)=F22(1:nm2) 
     b22(1:nm2)=F23(1:nm2) 
      
     h1(nm1)=F11(nm1) 
     h2(nm1)=F21(nm1) 
      
     b11(nm1)=F12(nm1) 
     b12(nm1)=0.d0 
     b21(nm1)=F22(nm1) 
     b22(nm1)=0.d0 
      
     h1(n)=(1.d0,0.d0) 
     h2(n)=0.d0   
     gamma1(1)=0.d0 
     gamma2(1)=(1.d0,0.d0) 
     g1(1)=(1.d0,0.d0) 
     g2(1)=0.d0 
      
     do k=1,nm2 
         
        auxMM11=gamma1(k)*F11(k)+gamma2(k)*F21(k) 
        auxMM12=gamma1(k)*F12(k)+gamma2(k)*F22(k) 
        auxMM13=gamma1(k)*F13(k)+gamma2(k)*F23(k) 
        auxMM21=F31(k) 
        auxMM22=F32(k) 
        auxMM23=F33(k) 
         
        kp1=k+1 
        MM11=V11(kp1)*auxMM11+V12(kp1)*auxMM21 
        MM12=V11(kp1)*auxMM12+V12(kp1)*auxMM22 
        MM13=V11(kp1)*auxMM13+V12(kp1)*auxMM23 
        MM22=V21(kp1)*auxMM12+V22(kp1)*auxMM22 
        MM23=V21(kp1)*auxMM13+V22(kp1)*auxMM23 
         
        sigma(k)=MM11 
        g1(kp1)=MM12 
        g2(kp1)=MM13 
        gamma1(kp1)=MM22 
        gamma2(kp1)=MM23 
     enddo 
      
     sigma(nm1)=(gamma1(nm1)*h1(nm1))+(gamma2(nm1)*h2(nm1)) 
     g1(n)=gamma1(nm1)*b11(nm1)+gamma2(nm1)*b21(nm1) 
     g2(n)=gamma1(nm1)*b12(nm1)+gamma2(nm1)*b22(nm1) 
      
      
     ! Compute beta (step 2 in Thm 3.1): 
      
     beta(1:2)=0.d0 
     beta(n)=p(n) 
     do k=nm1,3,-1 
        beta(k)=conjg(V11(k))*p(k)+conjg(V21(k))*beta(k+1) 
     enddo 
      
     ! Compute the vectors of diagonal and subdiagonal entries: 
      
     do k=1,n 
        diagonals(k)=g1(k)*h1(k)+g2(k)*h2(k)-p(k)*q(k) 
     enddo 
     do k=1,nm1 
        subdiag(k)=sigma(k)-p(k+1)*q(k) 
     enddo 
      
      
     ! Store available eigenvalues and look for first (from bottom) 
     ! negligible subdiagonal element: 
      
     repeat_flag=1 
     findm:do while (repeat_flag==1) 
        m=defl_size 
        smallsub:do while (m>1) 
           if (abs(subdiag(m))<threshold)then 
              !if (abs(subdiag(m))<(epsilon(1.d0)*abs(diagonals(m)+diagonals(m+1))))then 
              exit smallsub 
           end if 
           m=m-1 
        end do smallsub 
        if (defl_size==2) then 
           exit approx 
        end if 
        if (m==defl_size)then 
           !write(*,*)"deflation... defl_size = ",defl_size 
           myeig(defl_size+1)=diagonals(defl_size+1) ! one eigenvalue found 
           defl_size=m-1 
        else 
           repeat_flag=0 
        end if 
     end do findm 
      
     ! Consider the submatrix of indices m:defl_size+1 and look 
     ! for a place where the magic trick may happen... 
      
     alpha=diagonals(defl_size+1) ! shift 
     k=defl_size; ! now assume that the bulge-chasing process starts from k... 
     magic:do while (k>m) 
        x(1)=diagonals(k)-alpha 
        x(2)=subdiag(k) 
        call givens(x,Giv11,Giv12,Giv21,Giv22) 
        y(1)=Giv11*subdiag(k-1) 
        y(2)=Giv21*subdiag(k-1) 
        if (abs(y(2))<threshold) then 
           exit magic 
        end if 
        k=k-1 
     end do magic 
     m=k   
      
     ! From now on we will work on the submatrix of indices m:defl_size+1. 
      
     ! Increase iterations counter and set flag for random shift: 
     iterations(defl_size+1)=iterations(defl_size+1)+1 
     if (iterations(defl_size+1)>15) then 
        write(*,*)"random shift..." 
        random_flag=1 
        iterations(defl_size+1)=0
     end if 
      
     ! ****** BULGE CHASING *********************************** 
     ! 
     ! (step 3 in Thm. 3.1) 
     ! 
     ! Perform the bulge chasing process on the submatrix 
     ! (m:defl_size+1, m:defl_size+1). 
      
     qtau(m)=q(m) 
      
     ! Compute shift (alpha). 
     if (random_flag==1) then 
        call random_number(w1) 
        call random_number(w2) 
        alpha=(2.d0,0.d0)*w1-(1.d0, 0.d0)+(0.d0,1.d0)*w2-(0.d0, 1.d0) 
        random_flag=0 
     else 
        alpha=diagonals(defl_size+1) 
     end if 
      
     ! Compute and apply (on the left) the first Givens rotation: 
     x(1)=diagonals(m)-alpha 
     x(2)=subdiag(m) 
     call givens(x,Giv11,Giv12,Giv21,Giv22) 
     Q11(m)=conjg(Giv11) 
     Q12(m)=conjg(Giv21) 
     Q21(m)=conjg(Giv12) 
     Q22(m)=conjg(Giv22) 
     XX12=g1(m)*h1(m)+g2(m)*h2(m) 
     XX13=g1(m)*b11(m)+g2(m)*b21(m) 
     XX14=g1(m)*b12(m)+g2(m)*b22(m)     
      
     newp(m)=Giv11*p(m)+Giv12*p(m+1) 
     ptau(m+1)=Giv21*p(m)+Giv22*p(m+1) 
     sigmatau(m)=Giv21*XX12+Giv22*sigma(m) 
     gtau1(m+1)=Giv21*XX13+Giv22*g1(m+1) 
     gtau2(m+1)=Giv21*XX14+Giv22*g2(m+1) 
      
     ! ************** begin bulge chasing cycle ************* 
      
     k=m 
     chase:do while (k<defl_size) 
        ! Multiply by the old Giv on the right: 
         
        kp1=k+1 
        ! in the old way... 
        XXX11=sigmatau(k) 
        XXX12=gtau1(kp1)*h1(kp1)+gtau2(kp1)*h2(kp1) 
        XXX21=p(k+2)*qtau(k) 
        XXX22=sigma(kp1) 
        XXX31=qtau(k) 
        XXX32=q(kp1) 
        YYY12=XXX11*Q12(k)+XXX12*Q22(k) 
        YYY22=XXX21*Q12(k)+XXX22*Q22(k) 
        YYY31=XXX31*Q11(k)+XXX32*Q21(k) 
        YYY32=XXX31*Q12(k)+XXX32*Q22(k) 
        newq(k)=YYY31 
        qtau(kp1)=YYY32 
         
        ! ...and in the new way: 
        bulge11=sigmatau(k)-ptau(kp1)*qtau(k) 
        bulge12=gtau1(kp1)*h1(kp1)+gtau2(kp1)*h2(kp1)-ptau(kp1)*q(kp1) 
        bulge21=0.d0 
        bulge22=sigma(kp1)-p(k+2)*q(kp1) 
        ZZ11=bulge11*Q11(k)+bulge12*Q21(k) 
        ZZ12=bulge11*Q12(k)+bulge12*Q22(k) 
        ZZ21=bulge21*Q11(k)+bulge22*Q21(k) 
        ZZ22=bulge21*Q12(k)+bulge22*Q22(k) 
        bulge11=ZZ11 
        bulge12=ZZ12 
        bulge21=ZZ21 
        bulge22=ZZ22 
         
        ! Compute the new Giv: 
        x(1)=bulge11 
        x(2)=bulge21 
        call givens(x,Giv11,Giv12,Giv21,Giv22) 
        Q11(kp1)=conjg(Giv11) 
        Q12(kp1)=conjg(Giv21) 
        Q21(kp1)=conjg(Giv12) 
        Q22(kp1)=conjg(Giv22) 
         
        ! Multiply on the left by the new G: 
        x(1)=ptau(kp1) 
        x(2)=p(k+2) 
        y(1)=Giv11*x(1)+Giv12*x(2) 
        y(2)=Giv21*x(1)+Giv22*x(2) 
        newp(kp1)=y(1) 
        ptau(k+2)=y(2) 
        auxM11=YYY12 
        auxM12=gtau1(kp1)*b11(kp1)+gtau2(kp1)*b21(kp1) 
        auxM13=gtau1(kp1)*b12(kp1)+gtau2(kp1)*b22(kp1)   
         
        auxM21=YYY22 
        auxM22=g1(k+2) 
        auxM23=g2(k+2) 
        sigmatau(kp1)=Giv21*auxM11+Giv22*auxM21 
        gtau1(k+2)=Giv21*auxM12+Giv22*auxM22 
        gtau2(k+2)=Giv21*auxM13+Giv22*auxM23 
         
        ! Increase counter: 
        k=kp1      
         
     end do chase 
      
     ! Complete the update of p and q: 
     z(1)=qtau(defl_size) 
     z(2)=q(defl_size+1) 
     newq(defl_size)=z(1)*Q11(defl_size)+z(2)*Q21(defl_size) 
     newq(defl_size+1)=z(1)*Q12(defl_size)+z(2)*Q22(defl_size) 
      
     newp(defl_size+1)=ptau(defl_size+1) 
     newp(defl_size+2:n)=p(defl_size+2:n) 
     newq(defl_size+2:n)=q(defl_size+2:n) 
     newp(1:m-1)=p(1:m-1) 
     newq(1:m-1)=q(1:m-1) 
      
     ! **************** end bulge chasing cycle *************** 
      
      
     ! Update V and F (step 4 in Thm. 3.1), from index 1 to n=defl_size+1 
      
     ptau(1:m)=p(1:m)  
     x(1)=ptau(2) 
     x(2)=beta(3) 
     call givens(x,Giv11,Giv12,Giv21,Giv22) 
     Vtau11(2)=conjg(Giv11) 
     Vtau12(2)=conjg(Giv21) 
     Vtau21(2)=conjg(Giv12) 
     Vtau22(2)=conjg(Giv22) 
      
     y(1)=Giv11*x(1)+Giv12*x(2) 
     y(2)=Giv21*x(1)+Giv22*x(2) 
     auxA21=V11(2)*F21(1)+V12(2)*F31(1) 
     auxA22=V11(2)*F22(1)+V12(2)*F32(1) 
     auxA23=V11(2)*F23(1)+V12(2)*F33(1) 
     auxA31=V21(2)*F21(1)+V22(2)*F31(1)  
     auxA32=V21(2)*F22(1)+V22(2)*F32(1) 
     auxA33=V21(2)*F23(1)+V22(2)*F33(1)  
     auxB11=conjg(Q11(1)) 
     auxB12=conjg(Q21(1)) 
     auxB21=conjg(Q12(1)) 
     auxB22=conjg(Q22(1)) 
     auxC21=auxB21*F11(1)+auxB22*auxA21 
     auxC22=auxB21*F12(1)+auxB22*auxA22 
     auxC23=auxB21*F13(1)+auxB22*auxA23 
      
     Ftau11(1)=auxB11*F11(1)+auxB12*auxA21 
     Ftau12(1)=auxB11*F12(1)+auxB12*auxA22 
     Ftau13(1)=auxB11*F13(1)+auxB12*auxA23 
     Ftau21(1)=Giv11*auxC21+Giv12*auxA31 
     Ftau22(1)=Giv11*auxC22+Giv12*auxA32 
     Ftau23(1)=Giv11*auxC23+Giv12*auxA33 
     Ftau31(1)=Giv21*auxC21+Giv22*auxA31 
     Ftau32(1)=Giv21*auxC22+Giv22*auxA32 
     Ftau33(1)=Giv21*auxC23+Giv22*auxA33 
      
     if (defl_size>n-5) then 
        update:do k=1,n-3 
           kp1=k+1 
           kp2=k+2 
            
           auxB21=V11(kp2)*Vtau21(kp1) 
           auxB22=V11(kp2)*Vtau22(kp1) 
            
           Vhat11(kp1)=conjg(Q11(kp1))*Vtau11(kp1)+conjg(Q21(kp1))*auxB21 
           Vhat12(kp1)=conjg(Q11(kp1))*Vtau12(kp1)+conjg(Q21(kp1))*auxB22 
           Vhat13(kp1)=conjg(Q21(kp1))*V12(kp2) 
           Vhat21(kp1)=conjg(Q12(kp1))*Vtau11(kp1)+conjg(Q22(kp1))*auxB21 
           Vhat22(kp1)=conjg(Q12(kp1))*Vtau12(kp1)+conjg(Q22(kp1))*auxB22 
           Vhat23(kp1)=conjg(Q22(kp1))*V12(kp2) 
           Vhat31(kp1)=V21(kp2)*Vtau21(kp1) 
           Vhat32(kp1)=V21(kp2)*Vtau22(kp1) 
           Vhat33(kp1)=V22(kp2) 
            
           auxBB21=F11(kp1)*Q21(k) 
           auxBB22=F11(kp1)*Q22(k) 
           auxBB31=F21(kp1)*Q21(k) 
           auxBB32=F21(kp1)*Q22(k) 
           Fhat11(k)=Ftau11(k)*Q11(k)+Ftau12(k)*auxBB21+Ftau13(k)*auxBB31 
           Fhat12(k)=Ftau11(k)*Q12(k)+Ftau12(k)*auxBB22+Ftau13(k)*auxBB32 
           Fhat13(k)=Ftau12(k)*F12(kp1)+Ftau13(k)*F22(kp1) 
           Fhat14(k)=Ftau12(k)*F13(kp1)+Ftau13(k)*F23(kp1) 
           Fhat21(k)=Ftau21(k)*Q11(k)+Ftau22(k)*auxBB21+Ftau23(k)*auxBB31 
           Fhat22(k)=Ftau21(k)*Q12(k)+Ftau22(k)*auxBB22+Ftau23(k)*auxBB32 
           Fhat23(k)=Ftau22(k)*F12(kp1)+Ftau23(k)*F22(kp1) 
           Fhat24(k)=Ftau22(k)*F13(kp1)+Ftau23(k)*F23(kp1) 
           Fhat31(k)=Ftau31(k)*Q11(k)+Ftau32(k)*auxBB21+Ftau33(k)*auxBB31 
           Fhat32(k)=Ftau31(k)*Q12(k)+Ftau32(k)*auxBB22+Ftau33(k)*auxBB32 
           Fhat33(k)=Ftau32(k)*F12(kp1)+Ftau33(k)*F22(kp1) 
           Fhat34(k)=Ftau32(k)*F13(kp1)+Ftau33(k)*F23(kp1) 
           Fhat41(k)=F31(kp1)*Q21(k) 
           Fhat42(k)=F31(kp1)*Q22(k) 
           Fhat43(k)=F32(kp1) 
           Fhat44(k)=F33(kp1) 
            
           x(1)=conjg(Vhat12(kp1)) 
           x(2)=conjg(Vhat13(kp1)) 
            
           call givens(x,Giv11,Giv12,Giv21,Giv22) 
           ZZ11=conjg(Giv11) 
           ZZ12=conjg(Giv21) 
           ZZ21=conjg(Giv12) 
           ZZ22=conjg(Giv22) 
            
           ! WW(1,3)=0 
           WW11=Vhat11(kp1) 
           WW21=Vhat21(kp1) 
           WW31=Vhat31(kp1) 
           WW12=Vhat12(kp1)*ZZ11+Vhat13(kp1)*ZZ21 
           WW22=Vhat22(kp1)*ZZ11+Vhat23(kp1)*ZZ21 
           WW23=Vhat22(kp1)*ZZ12+Vhat23(kp1)*ZZ22 
           WW32=Vhat32(kp1)*ZZ11+Vhat33(kp1)*ZZ21 
           WW33=Vhat32(kp1)*ZZ12+Vhat33(kp1)*ZZ22 
            
           ! DD(4,1)=0 
           DD11=Fhat11(k) 
           DD12=Fhat12(k) 
           DD13=Fhat13(k) 
           DD14=Fhat14(k) 
           DD21=Fhat21(k) 
           DD22=Fhat22(k) 
           DD23=Fhat23(k) 
           DD24=Fhat24(k) 
           DD31=Giv11*Fhat31(k)+Giv12*Fhat41(k) 
           DD32=Giv11*Fhat32(k)+Giv12*Fhat42(k) 
           DD33=Giv11*Fhat33(k)+Giv12*Fhat43(k) 
           DD34=Giv11*Fhat34(k)+Giv12*Fhat44(k) 
           DD42=Giv21*Fhat32(k)+Giv22*Fhat42(k) 
           DD43=Giv21*Fhat33(k)+Giv22*Fhat43(k) 
           DD44=Giv21*Fhat34(k)+Giv22*Fhat44(k) 
            
           call split2(Vtau11(kp2),Vtau12(kp2),Vtau21(kp2),Vtau22(kp2),newV11(kp1),newV12(kp1),newV21(kp1),newV22(kp1),& 
           &WW11,WW12,WW21,WW22,WW23,WW31,WW32,WW33) 
            
           call split3wh(newF11(k),newF12(k),newF13(k),newF21(k),newF22(k),newF23(k),newF31(k),newF32(k),newF33(k),& 
           &Ftau11(kp1),Ftau12(kp1),Ftau13(kp1),Ftau21(kp1),Ftau22(kp1),Ftau23(kp1),Ftau31(kp1),Ftau32(kp1),Ftau33(kp1),& 
           &DD11,DD12,DD13,DD14,DD21,DD22,DD23,DD24,DD31,DD32,DD33,DD34,DD42,DD43,DD44) 
            
        end do update 
         
        ZZ11=conjg(Q11(nm1)) 
        ZZ12=conjg(Q21(nm1)) 
        ZZ21=conjg(Q12(nm1)) 
        ZZ22=conjg(Q22(nm1)) 
        newV11(nm1)=ZZ11*Vtau11(nm1)+ZZ12*Vtau21(nm1) 
        newV12(nm1)=ZZ11*Vtau12(nm1)+ZZ12*Vtau22(nm1) 
        newV21(nm1)=ZZ21*Vtau11(nm1)+ZZ22*Vtau21(nm1) 
        newV22(nm1)=ZZ21*Vtau12(nm1)+ZZ22*Vtau22(nm1)    
         
        auxA12=Q12(nm2)*Q11(nm1) 
        auxA13=Q12(nm2)*Q12(nm1) 
        auxA22=Q22(nm2)*Q11(nm1) 
        auxA23=Q22(nm2)*Q12(nm1) 
         
        newF11(nm2)=Ftau11(nm2)*Q11(nm2)+Ftau12(nm2)*Q21(nm2) 
        newF12(nm2)=Ftau11(nm2)*auxA12+Ftau12(nm2)*auxA22+Ftau13(nm2)*Q21(nm1) 
        newF13(nm2)=Ftau11(nm2)*auxA13+Ftau12(nm2)*auxA23+Ftau13(nm2)*Q22(nm1)  
        newF21(nm2)=Ftau21(nm2)*Q11(nm2)+Ftau22(nm2)*Q21(nm2) 
        newF22(nm2)=Ftau21(nm2)*auxA12+Ftau22(nm2)*auxA22+Ftau23(nm2)*Q21(nm1) 
        newF23(nm2)=Ftau21(nm2)*auxA13+Ftau22(nm2)*auxA23+Ftau23(nm2)*Q22(nm1)  
        newF31(nm2)=Ftau31(nm2)*Q11(nm2)+Ftau32(nm2)*Q21(nm2) 
        newF32(nm2)=Ftau31(nm2)*auxA12+Ftau32(nm2)*auxA22+Ftau33(nm2)*Q21(nm1) 
        newF33(nm2)=Ftau31(nm2)*auxA13+Ftau32(nm2)*auxA23+Ftau33(nm2)*Q22(nm1)  
         
        newF12(nm1)=0.d0 
        newF13(nm1)=0.d0 
        newF21(nm1)=0.d0 
        newF23(nm1)=0.d0 
        newF31(nm1)=0.d0 
        newF32(nm1)=0.d0 
        newF11(nm1)=(1.d0,0.d0) 
        newF22(nm1)=(1.d0,0.d0) 
        newF33(nm1)=(1.d0,0.d0) 
        newV11(1)=(1.d0,0.d0) 
        newV12(1)=0.d0 
        newV21(1)=0.d0 
        newV22(1)=(1.d0,0.d0) 
         
        p=newp 
        q=newq 
        V11=newV11 
        V12=newV12 
        V21=newV21  
        V22=newV22 
         
        F11=newF11 
        F33=newF33 
        F13=newF13 
        F21=newF21 
        F22=newF22 
        F23=newF23 
        F31=newF31 
        F12=newF12 
        F32=newF32 
         
     else 
        partialupdate:do k=1,defl_size+2 
           kp1=k+1 
           kp2=k+2 
            
           auxB21=V11(kp2)*Vtau21(kp1) 
           auxB22=V11(kp2)*Vtau22(kp1) 
            
           Vhat11(kp1)=conjg(Q11(kp1))*Vtau11(kp1)+conjg(Q21(kp1))*auxB21 
           Vhat12(kp1)=conjg(Q11(kp1))*Vtau12(kp1)+conjg(Q21(kp1))*auxB22 
           Vhat13(kp1)=conjg(Q21(kp1))*V12(kp2) 
           Vhat21(kp1)=conjg(Q12(kp1))*Vtau11(kp1)+conjg(Q22(kp1))*auxB21 
           Vhat22(kp1)=conjg(Q12(kp1))*Vtau12(kp1)+conjg(Q22(kp1))*auxB22 
           Vhat23(kp1)=conjg(Q22(kp1))*V12(kp2) 
           Vhat31(kp1)=V21(kp2)*Vtau21(kp1) 
           Vhat32(kp1)=V21(kp2)*Vtau22(kp1) 
           Vhat33(kp1)=V22(kp2) 
            
            
           auxBB21=F11(kp1)*Q21(k) 
           auxBB22=F11(kp1)*Q22(k) 
           auxBB31=F21(kp1)*Q21(k) 
           auxBB32=F21(kp1)*Q22(k) 
            
           Fhat11(k)=Ftau11(k)*Q11(k)+Ftau12(k)*auxBB21+Ftau13(k)*auxBB31 
           Fhat12(k)=Ftau11(k)*Q12(k)+Ftau12(k)*auxBB22+Ftau13(k)*auxBB32 
           Fhat13(k)=Ftau12(k)*F12(kp1)+Ftau13(k)*F22(kp1) 
           Fhat14(k)=Ftau12(k)*F13(kp1)+Ftau13(k)*F23(kp1) 
           Fhat21(k)=Ftau21(k)*Q11(k)+Ftau22(k)*auxBB21+Ftau23(k)*auxBB31 
           Fhat22(k)=Ftau21(k)*Q12(k)+Ftau22(k)*auxBB22+Ftau23(k)*auxBB32 
           Fhat23(k)=Ftau22(k)*F12(kp1)+Ftau23(k)*F22(kp1) 
           Fhat24(k)=Ftau22(k)*F13(kp1)+Ftau23(k)*F23(kp1) 
           Fhat31(k)=Ftau31(k)*Q11(k)+Ftau32(k)*auxBB21+Ftau33(k)*auxBB31 
           Fhat32(k)=Ftau31(k)*Q12(k)+Ftau32(k)*auxBB22+Ftau33(k)*auxBB32 
           Fhat33(k)=Ftau32(k)*F12(kp1)+Ftau33(k)*F22(kp1) 
           Fhat34(k)=Ftau32(k)*F13(kp1)+Ftau33(k)*F23(kp1) 
           Fhat41(k)=F31(kp1)*Q21(k) 
           Fhat42(k)=F31(kp1)*Q22(k) 
           Fhat43(k)=F32(kp1) 
           Fhat44(k)=F33(kp1) 
            
           x(1)=conjg(Vhat12(kp1)) 
           x(2)=conjg(Vhat13(kp1)) 
            
           call givens(x,Giv11,Giv12,Giv21,Giv22) 
           ZZ11=conjg(Giv11) 
           ZZ12=conjg(Giv21) 
           ZZ21=conjg(Giv12) 
           ZZ22=conjg(Giv22) 
            
           ! WW(1,3)=0 
           WW11=Vhat11(kp1) 
           WW21=Vhat21(kp1) 
           WW31=Vhat31(kp1) 
           WW12=Vhat12(kp1)*ZZ11+Vhat13(kp1)*ZZ21 
           WW22=Vhat22(kp1)*ZZ11+Vhat23(kp1)*ZZ21 
           WW23=Vhat22(kp1)*ZZ12+Vhat23(kp1)*ZZ22 
           WW32=Vhat32(kp1)*ZZ11+Vhat33(kp1)*ZZ21 
           WW33=Vhat32(kp1)*ZZ12+Vhat33(kp1)*ZZ22 
            
           ! DD(4,1)=0 
           DD11=Fhat11(k) 
           DD12=Fhat12(k) 
           DD13=Fhat13(k) 
           DD14=Fhat14(k) 
           DD21=Fhat21(k) 
           DD22=Fhat22(k) 
           DD23=Fhat23(k) 
           DD24=Fhat24(k) 
           DD31=Giv11*Fhat31(k)+Giv12*Fhat41(k) 
           DD32=Giv11*Fhat32(k)+Giv12*Fhat42(k) 
           DD33=Giv11*Fhat33(k)+Giv12*Fhat43(k) 
           DD34=Giv11*Fhat34(k)+Giv12*Fhat44(k) 
           DD42=Giv21*Fhat32(k)+Giv22*Fhat42(k) 
           DD43=Giv21*Fhat33(k)+Giv22*Fhat43(k) 
           DD44=Giv21*Fhat34(k)+Giv22*Fhat44(k) 
            
           call split2(Vtau11(kp2),Vtau12(kp2),Vtau21(kp2),Vtau22(kp2),newV11(kp1),newV12(kp1),newV21(kp1),newV22(kp1),& 
           &WW11,WW12,WW21,WW22,WW23,WW31,WW32,WW33) 
            
           call split3wh(newF11(k),newF12(k),newF13(k),newF21(k),newF22(k),newF23(k),newF31(k),newF32(k),newF33(k),& 
           &Ftau11(kp1),Ftau12(kp1),Ftau13(kp1),Ftau21(kp1),Ftau22(kp1),Ftau23(kp1),Ftau31(kp1),Ftau32(kp1),Ftau33(kp1),& 
           &DD11,DD12,DD13,DD14,DD21,DD22,DD23,DD24,DD31,DD32,DD33,DD34,DD42,DD43,DD44) 
            
        end do partialupdate 
        p=newp 
        q=newq 
        V11(1:defl_size+3)=newV11(1:defl_size+3) 
        V12(1:defl_size+3)=newV12(1:defl_size+3) 
        V21(1:defl_size+3)=newV21(1:defl_size+3)  
        V22(1:defl_size+3)=newV22(1:defl_size+3) 
         
        V11(defl_size+4)=Vtau11(defl_size+4) 
        V12(defl_size+4)=Vtau12(defl_size+4) 
        V21(defl_size+4)=Vtau21(defl_size+4) 
        V22(defl_size+4)=Vtau22(defl_size+4) 
         
        F11(1:defl_size+3)=newF11(1:defl_size+3) 
        F12(1:defl_size+3)=newF12(1:defl_size+3) 
        F13(1:defl_size+3)=newF13(1:defl_size+3) 
        F21(1:defl_size+3)=newF21(1:defl_size+3) 
        F22(1:defl_size+3)=newF22(1:defl_size+3) 
        F23(1:defl_size+3)=newF23(1:defl_size+3) 
        F31(1:defl_size+3)=newF31(1:defl_size+3) 
        F32(1:defl_size+3)=newF32(1:defl_size+3) 
        F33(1:defl_size+3)=newF33(1:defl_size+3) 
         
     end if 
      
  end do approx 

CALL SYSTEM_CLOCK(COUNT=clock_end)  
  our_time=(REAL(clock_end - clock_start)/clock_rate) 
  write(*,*)"time (quasiseparable approximation) = ",our_time 
   
  ! ************ end approx cycle ********************** 
   
   
  ! In order to compute the first LAPACKSIZE eigenvalues, we 
  ! recover the matrix (stored in auxAA) and use Lapack  
CALL SYSTEM_CLOCK(COUNT_RATE=clock_rate) 
  CALL SYSTEM_CLOCK(COUNT=clock_start) 

  h1(1:nm2)=F11(1:nm2) 
  h2(1:nm2)=F21(1:nm2)  
   
  b11(1:nm2)=F12(1:nm2) 
  b12(1:nm2)=F13(1:nm2) 
  b21(1:nm2)=F22(1:nm2) 
  b22(1:nm2)=F23(1:nm2) 
   
  h1(nm1)=F11(nm1) 
  h2(nm1)=F21(nm1) 
   
  b11(nm1)=F12(nm1) 
  b12(nm1)=0.d0 
  b21(nm1)=F22(nm1) 
  b22(nm1)=0.d0 
   
  h1(n)=(1.d0,0.d0) 
  h2(n)=0.d0   
  gamma1(1)=0.d0 
  gamma2(1)=(1.d0,0.d0) 
  g1(1)=(1.d0,0.d0) 
  g2(1)=0.d0 
   
  do k=1,nm2 
      
     auxMM11=gamma1(k)*F11(k)+gamma2(k)*F21(k) 
     auxMM12=gamma1(k)*F12(k)+gamma2(k)*F22(k) 
     auxMM13=gamma1(k)*F13(k)+gamma2(k)*F23(k) 
     auxMM21=F31(k) 
     auxMM22=F32(k) 
     auxMM23=F33(k) 
      
     kp1=k+1 
     MM11=V11(kp1)*auxMM11+V12(kp1)*auxMM21 
     MM12=V11(kp1)*auxMM12+V12(kp1)*auxMM22 
     MM13=V11(kp1)*auxMM13+V12(kp1)*auxMM23 
     MM22=V21(kp1)*auxMM12+V22(kp1)*auxMM22 
     MM23=V21(kp1)*auxMM13+V22(kp1)*auxMM23 
      
     sigma(k)=MM11 
     g1(kp1)=MM12 
     g2(kp1)=MM13 
     gamma1(kp1)=MM22 
     gamma2(kp1)=MM23 
  enddo 
  sigma(nm1)=(gamma1(nm1)*h1(nm1))+(gamma2(nm1)*h2(nm1)) 
  g1(n)=gamma1(nm1)*b11(nm1)+gamma2(nm1)*b21(nm1) 
  g2(n)=gamma1(nm1)*b12(nm1)+gamma2(nm1)*b22(nm1) 
   
   
  ! Compute beta (step 2 in Thm 3.1): 
   
  beta(1:2)=0.d0 
  beta(n)=p(n) 
  do k=nm1,3,-1 
     beta(k)=conjg(V11(k))*p(k)+conjg(V21(k))*beta(k+1) 
  enddo 
   
  ! Compute the vectors of diagonal and subdiagonal entries: 
   
  do k=1,n 
     diagonals(k)=g1(k)*h1(k)+g2(k)*h2(k)-p(k)*q(k) 
  enddo 
  do k=1,nm1 
     subdiag(k)=sigma(k)-p(k+1)*q(k) 
  enddo 

  allocate(auxAA(lapacksize,lapacksize))

  ! store diagonal entries
  do j=1,lapacksize
     !auxAA(j,j)=g1(j)*h1(j)+g2(j)*h2(j)-p(j)*q(j)
     auxAA(j,j)=diagonals(j)
  end do
  ! store subdiagonal entries
  do j=1,lapacksize-1
     !auxAA(j+1,j)=sigma(j)-p(j+1)*q(j)
     auxAA(j+1,j)=subdiag(j)
  end do
  ! set zero entries
  do i=3,lapacksize
     do j=1,i-2
        auxAA(i,j)=0.d0
     end do
  end do
  ! store entries in upper triangular part

  do j=lapacksize,2,-1
     y(1)=h1(j)
     y(2)=h2(j)
     do i=(j-1),1,-1
        x(1)=b11(i)*y(1)+b12(i)*y(2) 
        x(2)=b21(i)*y(1)+b22(i)*y(2) 
        auxAA(i,j)=g1(i)*x(1)+g2(i)*x(2)-p(i)*q(j)
        y(1)=x(1)
        y(2)=x(2)
     end do
      
  end do

  allocate(vl(1,lapacksize),vr(1,lapacksize),eig(lapacksize),rwork(2*lapacksize))
  LWORK = 6*lapacksize+100 
  allocate (WORK(LWORK)) 
  call ZGEEV('n','n',lapacksize,auxAA,lapacksize,eig,vL,1,vR,1,WORK,LWORK,RWORK,INFO) 
  myeig(1:lapacksize)=eig 

CALL SYSTEM_CLOCK(COUNT=clock_end)  
  our_time=(REAL(clock_end - clock_start)/clock_rate) 
  write(*,*)"time (call to Lapack) = ",our_time 
  !deallocate(vL,vR,eig,work,rwork) 
  if (info.ne.0) then
     write(*,*)"Lapack does not converge"
  end if
  return
end subroutine begg_eig_lapack


 
subroutine givens(x,G11,G12,G21,G22) 
  ! Given x=[x1; x2], computes the 2x2 Givens matrix G
  ! such that the second entry in G*x is zero.
  implicit none 
  complex(kind(1d0)), intent(in)  :: x(2) 
  complex(kind(1d0)), intent(out) :: G11,G12,G21,G22 
  real(kind(1d0)) :: s,w1,w2 
  w1=abs(x(1)) 
  w2=abs(x(2)) 
  G12=(0.d0,0.d0) 
  G21=(0.d0,0.d0) 
  G11=(1.d0,0.d0) 
  G22=(1.d0,0.d0) 
  if(max(w1,w2)<3.d-16)return 
   
  s=1/sqrt(w1*w1+w2*w2) 
  G11= conjg(s*x(1)) 
  G12= conjg(s*x(2)) 
  G21=-s*x(2) 
  G22= s*x(1) 
  return 
end subroutine givens 
 
subroutine split2(A11,A12,A21,A22,B11,B12,B21,B22,W11,W12,W21,W22,W23,W31,W32,W33) 
  ! factorizes W 
  implicit none 
  complex(kind(1d0)), intent(in)  :: W11,W12,W21,W22,W23,W31,W32,W33 
  complex(kind(1d0)), intent(out) :: A11,A12,A21,A22,B11,B12,B21,B22 
  complex(kind(1d0)) :: z2(2),z3(2) 
  real(kind(1d0)) :: norm2,norm3 
  B11=W11 
  B12=W12 
  z2(1)=W21 
  z2(2)=W22 
  z3(1)=W31 
  z3(2)=W32 
   
  norm2=sqrt(abs(dot_product(z2,z2))) 
  norm3=sqrt(abs(dot_product(z3,z3))) 
  if (norm2<norm3) then 
     B21=z3(1)/norm3 
     B22=z3(2)/norm3 
  else 
     B21=z2(1)/norm2 
     B22=z2(2)/norm2 
  end if 
  A11=W21*conjg(B21)+W22*conjg(B22) 
  A12=W23 
  A21=W31*conjg(B21)+W32*conjg(B22) 
  A22=W33 
  return 
end subroutine split2 
 
subroutine split3wh(A11,A12,A13,A21,A22,A23,A31,A32,A33,B11,B12,B13,B21,B22,B23,B31,B32,B33,& 
  &D11,D12,D13,D14,D21,D22,D23,D24,D31,D32,D33,D34,D42,D43,D44) 
  ! factorizes D 
  implicit none 
  complex(kind(1d0)), intent(in)  :: D11,D12,D13,D14,D21,D22,D23,D24,D31,D32,D33,D34,D42,D43,D44 
  complex(kind(1d0)), intent(out) :: A11,A12,A13,A21,A22,A23,A31,A32,A33,B11,B12,B13,B21,B22,B23,B31,B32,B33 
  complex(kind(1d0)) :: v(3),H11,H12,H13,H21,H22,H23,H32,H33,M22,M23,M24,M32,M33,M34 
  v(1)=D11 
  v(2)=D21 
  v(3)=D31 
  call householder(H11,H12,H13,H21,H22,H23,H32,H33,v)   
  M22=H21*D12+H22*D22+H23*D32 
  M23=H21*D13+H22*D23+H23*D33 
  M24=H21*D14+H22*D24+H23*D34 
  M32=H32*D22+H33*D32 
  M33=H32*D23+H33*D33 
  M34=H32*D24+H33*D34 
  B11=M22 
  B12=M23 
  B13=M24 
  B21=M32 
  B22=M33 
  B23=M34 
  B31=D42 
  B32=D43 
  B33=D44   
  A11=conjg(H11) 
  A12=conjg(H21) 
  A13=0.d0 
  A21=conjg(H12) 
  A22=conjg(H22) 
  A23=conjg(H32) 
  A31=conjg(H13) 
  A32=conjg(H23) 
  A33=conjg(H33) 
  return 
end subroutine split3wh 
 
subroutine householder(A11,A12,A13,A21,A22,A23,A32,A33,x) 
  ! computes A such that A*x=[norm(x); 0; 0]. 
  implicit none 
  complex(kind(1d0)), intent(in)  :: x(3) 
  complex(kind(1d0)), intent(out) :: A11,A12,A13,A21,A22,A23,A32,A33 
  complex(kind(1d0)) :: v(3),Giv11,Giv12,Giv21,Giv22,G11,G12,G21,G22 
  v=x 
  call givens(v(2:3),Giv11,Giv12,Giv21,Giv22) 
  v(2)=Giv11*x(2)+Giv12*x(3) 
  call givens(v(1:2),G11,G12,G21,G22) 
  A11=G11 
  A12=G12*Giv11 
  A13=G12*Giv12 
  A21=G21 
  A22=G22*Giv11 
  A23=G22*Giv12 
  !A31=0.d0 
  A32=Giv21 
  A33=Giv22 
  return 
end subroutine householder 
 
